import math
import torch
import torch.nn as nn
from collections import OrderedDict
import ltr.models.target_classifier.linear_filter as target_clf
import ltr.models.target_classifier.features as clf_features
import ltr.models.target_classifier.initializer as clf_initializer
import ltr.models.target_classifier.optimizer as clf_optimizer
import ltr.models.bbreg as bbmodels
import ltr.models.backbone as backbones
from ltr import model_constructor


class DiMPnet(nn.Module):
    """The DiMP network.
    args:
        feature_extractor:  Backbone feature extractor network. Must return a dict of feature maps
        classifier:  Target classification module.
        bb_regressor:  Bounding box regression module.
        classification_layer:  Name of the backbone feature layer to use for classification.
        bb_regressor_layer:  Names of the backbone layers to use for bounding box regression.
        train_feature_extractor:  Whether feature extractor should be trained or not."""

    # classifier2
    def __init__(self, feature_extractor, classifier1, bb_regressor, classification_layer, bb_regressor_layer,
                 train_feature_extractor=True):
        super().__init__()

        self.feature_extractor = feature_extractor
        self.classifier = classifier1
        self.bb_regressor = bb_regressor
        self.classification_layer = [classification_layer] if isinstance(classification_layer,
                                                                         str) else classification_layer
        self.bb_regressor_layer = bb_regressor_layer
        self.output_layers = sorted(list(set(self.classification_layer + self.bb_regressor_layer)))

        if not train_feature_extractor:
            for p in self.feature_extractor.parameters():
                p.requires_grad_(False)

        ##################################################################################
        """
        self.smooth012 = nn.Conv2d(1024, 1024, kernel_size=3, stride=1, padding=1)
        self.smooth34 = nn.Conv2d(1024, 1024, kernel_size=3, stride=1, padding=1)
        self.smooth5 = nn.Conv2d(1024, 1024, kernel_size=3, stride=1, padding=1)

        self.smooth = [self.smooth012, self.smooth34, self.smooth5]
        for m in self.smooth:
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
                if m.bias is not None:
                    m.bias.data.zero_()
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()
        """

    def forward(self, train_imgs, test_imgs, train_bb, test_proposals, *args, **kwargs):
        """Runs the DiMP network the way it is applied during training.
        The forward function is ONLY used for training. Call the individual functions during tracking.
        args:
            train_imgs:  Train image samples (images, sequences, 3, H, W).
            test_imgs:  Test image samples (images, sequences, 3, H, W).
            trian_bb:  Target boxes (x,y,w,h) for the train images. Dims (images, sequences, 4).
            test_proposals:  Proposal boxes to use for the IoUNet (bb_regressor) module.
            *args, **kwargs:  These are passed to the classifier module.
        returns:
            test_scores:  Classification scores on the test samples.
            iou_pred:  Predicted IoU scores for the test_proposals."""

        assert train_imgs.dim() == 5 and test_imgs.dim() == 5, 'Expect 5 dimensional inputs'

        # Extract backbone features
        train_feat = self.extract_backbone_features(train_imgs.view(-1, *train_imgs.shape[-3:]))
        test_feat = self.extract_backbone_features(test_imgs.view(-1, *test_imgs.shape[-3:]))
        """
        train_feat = test_feat = 
        dict(3) layer2: [15, 512, 36, 36], layer3: [15, 1024, 18, 18], layer3_all6: [15, 1024, 18, 18]
        """
        ####################################################################################################
        layer3_all6_train = train_feat['layer3_all6']
        layer3_all6_test = test_feat['layer3_all6']

        self.second_filter_train_feat = layer3_all6_train[1]
        self.second_filter_test_feat = layer3_all6_test[1]

        train_feat_clf = layer3_all6_train[5]
        test_feat_clf = layer3_all6_test[5]

        # Run classifier module ##############################################
        target_scores = self.classifier(train_feat_clf, test_feat_clf, train_bb, *args, **kwargs)

        # Get bb_regressor features
        train_feat_iou = self.get_backbone_bbreg_feat(train_feat)
        test_feat_iou = self.get_backbone_bbreg_feat(test_feat)

        # Run the IoUNet module
        iou_pred = self.bb_regressor(train_feat_iou, test_feat_iou, train_bb, test_proposals)

        # print(str(target_scores_1[0].shape))
        """
        train_feat = dict() layer2: [15, 512, 36, 36], layer3: [15, 1024, 18, 18]

        train_imgs = [3, 5, 3, 288, 288]
        test_imgs = [3, 5, 3, 288, 288]

        target_scores_1 = list[6]: [3, 5 ,19, 19]
        iou_pred = [3, 5, 8]
        """
        return target_scores, iou_pred

    def second_filter_feat(self):
        return self.second_filter_train_feat, self.second_filter_test_feat

    def get_backbone_clf_feat(self, backbone_feat):
        """
        返回单层特征，layer3 是一个tensor【23， 1024， 14， 14】
        """
        feat = OrderedDict({l: backbone_feat[l] for l in self.classification_layer})
        if len(self.classification_layer) == 1:
            return feat[self.classification_layer[0]]
        # return feat['layer2'], feat['layer3']
        return feat

    def get_backbone_bbreg_feat(self, backbone_feat):
        return [backbone_feat[l] for l in self.bb_regressor_layer]

    def extract_classification_feat(self, init_backbone_feat_iou, init_backbone_feat_secondF):

        layer3_cls_feat_pretrained = self.classifier.extract_classification_feat(init_backbone_feat_iou['layer3'])
        layer3_cls_feat_secondF = self.classifier.extract_classification_feat(init_backbone_feat_secondF['layer3_all6'][1])

        return layer3_cls_feat_secondF, layer3_cls_feat_pretrained

    def extract_backbone_features(self, im, layers=None):
        if layers is None:
            layers = self.output_layers
        return self.feature_extractor(im, layers)

    def extract_features(self, im, layers=None):
        if layers is None:
            layers = self.bb_regressor_layer + ['classification']
        if 'classification' not in layers:
            return self.feature_extractor(im, layers)
        backbone_layers = sorted(list(set([l for l in layers + self.classification_layer if l != 'classification'])))
        all_feat = self.feature_extractor(im, backbone_layers)
        all_feat['classification'] = self.extract_classification_feat(all_feat)
        return OrderedDict({l: all_feat[l] for l in layers})


class SecondFilter(nn.Module):
    def __init__(self, dimp_pretrained, classifier, backbone_net_second_filter):
        super().__init__()

        self.dimp_pretrained = dimp_pretrained
        self.classifier = classifier
        self.backbone_net_second_filter = backbone_net_second_filter

    def forward(self, train_imgs, test_imgs, train_bb, test_proposals, *args, **kwargs):
        target_scores_pretrained, iou_pred = self.dimp_pretrained(train_imgs, test_imgs, train_bb, test_proposals)

        train_feat =  self.backbone_net_second_filter(train_imgs.view(-1, *train_imgs.shape[-3:]), ['layer3_all6'])
        test_feat =  self.backbone_net_second_filter(test_imgs.view(-1, *train_imgs.shape[-3:]), ['layer3_all6'])

        train_feat_clf = train_feat['layer3_all6'][1]
        test_feat_clf = test_feat['layer3_all6'][1]

        target_scores_second = self.classifier(train_feat_clf, test_feat_clf, train_bb, *args, **kwargs)

        return target_scores_pretrained, target_scores_second, iou_pred

    def get_backbone_clf_feat(self, backbone_feat):
        """
        返回单层特征，layer3 是一个tensor【23， 1024， 14， 14】
        """
        feat = OrderedDict({l: backbone_feat[l] for l in self.classification_layer})
        if len(self.dimp_pretrained.classification_layer) == 1:
            return feat[self.dimp_pretrained.classification_layer[0]]
        # return feat['layer2'], feat['layer3']
        return feat

    def get_backbone_bbreg_feat(self, backbone_feat):
        return [backbone_feat[l] for l in self.dimp_pretrained.bb_regressor_layer]

    def extract_classification_feat(self, init_backbone_feat_iou, init_backbone_feat_secondF):
        layer3_pretrain = init_backbone_feat_iou['layer3']
        layer3_all6_2 = init_backbone_feat_secondF['layer3_all6']

        layer3_final_2 = layer3_all6_2[1]

        layer3_cls_feat_secondF = self.classifier.extract_classification_feat(layer3_final_2)
        layer3_cls_feat_pretrain = self.dimp_pretrained.classifier.extract_classification_feat(layer3_pretrain)

        return layer3_cls_feat_secondF, layer3_cls_feat_pretrain

    def extract_backbone_features(self, im, layers=None):
        secondF_backbone_feature = self.backbone_net_second_filter(im,[ 'layer3_all6'] )
        iou_first_backbone_feature = self.dimp_pretrained.extract_backbone_features(im, ['layer2', 'layer3'])
        return iou_first_backbone_feature, secondF_backbone_feature

    def extract_features(self, im, layers=None):
        if layers is None:
            layers = self.dimp_pretrained.bb_regressor_layer + ['classification']
        if 'classification' not in layers:
            return self.dimp_pretrained.feature_extractor(im, layers)
        backbone_layers = sorted(list(set([l for l in layers + self.classification_layer if l != 'classification'])))
        all_feat = self.dimp_pretrained.feature_extractor(im, backbone_layers)
        all_feat['classification'] = self.dimp_pretrained.extract_classification_feat(all_feat)
        return OrderedDict({l: all_feat[l] for l in layers})


@model_constructor
def dimpnet18(filter_size=1, optim_iter=5, optim_init_step=1.0, optim_init_reg=0.01,
              classification_layer='layer3', feat_stride=16, backbone_pretrained=True, clf_feat_blocks=1,
              clf_feat_norm=True, init_filter_norm=False, final_conv=True,
              out_feature_dim=256, init_gauss_sigma=1.0, num_dist_bins=5, bin_displacement=1.0,
              mask_init_factor=4.0, iou_input_dim=(256, 256), iou_inter_dim=(256, 256),
              score_act='relu', act_param=None, target_mask_act='sigmoid',
              detach_length=float('Inf')):
    # Backbone
    backbone_net = backbones.resnet18(pretrained=backbone_pretrained)

    # Feature normalization
    norm_scale = math.sqrt(1.0 / (out_feature_dim * filter_size * filter_size))

    # Classifier features
    clf_feature_extractor = clf_features.residual_basic_block(num_blocks=clf_feat_blocks, l2norm=clf_feat_norm,
                                                              final_conv=final_conv, norm_scale=norm_scale,
                                                              out_dim=out_feature_dim)

    # Initializer for the DiMP classifier
    initializer = clf_initializer.FilterInitializerLinear(filter_size=filter_size, filter_norm=init_filter_norm,
                                                          feature_dim=out_feature_dim)

    # Optimizer for the DiMP classifier
    optimizer = clf_optimizer.DiMPSteepestDescentGN(num_iter=optim_iter, feat_stride=feat_stride,
                                                    init_step_length=optim_init_step,
                                                    init_filter_reg=optim_init_reg, init_gauss_sigma=init_gauss_sigma,
                                                    num_dist_bins=num_dist_bins,
                                                    bin_displacement=bin_displacement,
                                                    mask_init_factor=mask_init_factor,
                                                    score_act=score_act, act_param=act_param, mask_act=target_mask_act,
                                                    detach_length=detach_length)

    # The classifier module
    classifier = target_clf.LinearFilter(filter_size=filter_size, filter_initializer=initializer,
                                         filter_optimizer=optimizer, feature_extractor=clf_feature_extractor)

    # Bounding box regressor
    bb_regressor = bbmodels.AtomIoUNet(pred_input_dim=iou_input_dim, pred_inter_dim=iou_inter_dim)

    # DiMP network
    net = DiMPnet(feature_extractor=backbone_net, classifier=classifier, bb_regressor=bb_regressor,
                  classification_layer=classification_layer, bb_regressor_layer=['layer2', 'layer3'])
    return net


@model_constructor  # test_feat_clf
def dimpnet50(filter_size=1, optim_iter=5, optim_init_step=1.0, optim_init_reg=0.01,
              classification_layer=['layer3', 'layer3_all6'], feat_stride=16, backbone_pretrained=True,
              clf_feat_blocks=0,
              clf_feat_norm=True, init_filter_norm=False, final_conv=True,
              out_feature_dim=512, init_gauss_sigma=1.0, num_dist_bins=5, bin_displacement=1.0,
              mask_init_factor=4.0, iou_input_dim=(256, 256), iou_inter_dim=(256, 256),
              score_act='relu', act_param=None, target_mask_act='sigmoid',
              detach_length=float('Inf')):
    # Backbone
    backbone_net = backbones.resnet50(pretrained=backbone_pretrained)
    backbone_net_second_filter = backbones.resnet50(pretrained=backbone_pretrained)

    # Feature normalization
    norm_scale = math.sqrt(1.0 / (out_feature_dim * filter_size * filter_size))

    # Classifier features
    """
    layer3: 
    clf_feat_blocks = 0
    clf_feat_norm = True
    final_conv = True
    out_feature_dim = 1024/2 = 512
    """
    clf_feature_extractor = clf_features.residual_bottleneck(num_blocks=clf_feat_blocks, l2norm=clf_feat_norm,
                                                             final_conv=final_conv, norm_scale=norm_scale,
                                                             out_dim=out_feature_dim)
    clf_feature_extractor2 = clf_features.residual_bottleneck(num_blocks=clf_feat_blocks, l2norm=clf_feat_norm,
                                                             final_conv=final_conv, norm_scale=norm_scale,
                                                             out_dim=out_feature_dim)
    """
    feature_dim=256, num_blocks=1, l2norm=True, final_conv=False, norm_scale=1.0, out_dim=None,
                        interp_cat=False
    """
    """
    layer3: 
    filter_size = 4
    init_filter_norm = False
    out_feature_dim = 1024/2 = 512
    """
    # Initializer for the DiMP classifier
    initializer = clf_initializer.FilterInitializerLinear(filter_size=filter_size, filter_norm=init_filter_norm,
                                                          feature_dim=out_feature_dim)
    initializer2 = clf_initializer.FilterInitializerLinear(filter_size=filter_size, filter_norm=init_filter_norm,
                                                          feature_dim=out_feature_dim)
    # filter_size=1, feature_dim=256, feature_stride=16, pool_square=False, filter_norm=True, conv_ksz=3
    """
    ilter_size=1, feature_dim=256, feature_stride=16, pool_square=False, filter_norm=True,
    num_filter_pre_convs=1, num_filter_post_convs=0
    """
    """
    layer3: 
    optim_iter = 5
    feat_stride = 16 
    optim_init_step = 0.9
    optim_init_reg = 0.1 init_gauss_sigma = 0.9 num_dist_bins = 100 mask_init_factor=3.0
    score_act = relu act_param=none detach_length=inf
    """
    """
     num_iter=1, feat_stride=16, init_step_length=1.0,
                 init_filter_reg=1e-2, init_gauss_sigma=1.0, num_dist_bins=5, bin_displacement=1.0, mask_init_factor=4.0,
                 score_act='relu', act_param=None, min_filter_reg=1e-3, mask_act='sigmoid',
                 detach_length=float('Inf')
    """
    # Optimizer for the DiMP classifier
    optimizer = clf_optimizer.DiMPSteepestDescentGN(num_iter=optim_iter, feat_stride=feat_stride,
                                                    init_step_length=optim_init_step,
                                                    init_filter_reg=optim_init_reg, init_gauss_sigma=init_gauss_sigma,
                                                    num_dist_bins=num_dist_bins,
                                                    bin_displacement=bin_displacement,
                                                    mask_init_factor=mask_init_factor,
                                                    score_act=score_act, act_param=act_param, mask_act=target_mask_act,
                                                    detach_length=detach_length)
    optimizer2 = clf_optimizer.DiMPSteepestDescentGN(num_iter=optim_iter, feat_stride=feat_stride,
                                                    init_step_length=optim_init_step,
                                                    init_filter_reg=optim_init_reg, init_gauss_sigma=init_gauss_sigma,
                                                    num_dist_bins=num_dist_bins,
                                                    bin_displacement=bin_displacement,
                                                    mask_init_factor=mask_init_factor,
                                                    score_act=score_act, act_param=act_param, mask_act=target_mask_act,
                                                    detach_length=detach_length)
    """
    layer3: 
    filter_size = 4
    initializer = FilterInitializerLinear
    optimizer = DiMPSteepestDesecentGN
    clf_feature_extractor = clf_feature_extractor for layer 3
    """
    # The classifier module
    """
    filter_size, filter_initializer, filter_optimizer=None, feature_extractor=None
    """
    classifier = target_clf.LinearFilter(filter_size=filter_size, filter_initializer=initializer,
                                         filter_optimizer=optimizer, feature_extractor=clf_feature_extractor)
    classifier2 = target_clf.LinearFilter(filter_size=filter_size, filter_initializer=initializer2,
                                         filter_optimizer=optimizer2, feature_extractor=clf_feature_extractor2)
    # Bounding box regressor
    bb_regressor = bbmodels.AtomIoUNet(input_dim=(4 * 128, 4 * 256), pred_input_dim=iou_input_dim,
                                       pred_inter_dim=iou_inter_dim)

    # DiMP network
    DiMPnet_pretrained = DiMPnet(feature_extractor=backbone_net, classifier1=classifier, bb_regressor=bb_regressor,
                  classification_layer=classification_layer, bb_regressor_layer=['layer2', 'layer3'])

    net = SecondFilter(DiMPnet_pretrained, classifier2, backbone_net_second_filter)
    return net
